// debugger

#include "OurTypes.h"
#include "pilotcpu\pilotcpu.h"

const int kcpszArgsMax = 50;
const int kcchSymbolMax = 50;

typedef ulong LocalID;

typedef DWORD VirtualAddress;   // va
const VirtualAddress kvaInvalid = (VirtualAddress)-1;
const VirtualAddress kvaNextAppStartup = 0x10c0c5b8;

// NOTE: The VariableList class is probably temporary. If any other decent
// collection classes are used then they can replace it.

struct Variable {
    struct Variable *pvarNext;
    char *pszName;
    char *pszValue;
};

class VariableList {
public:
    VariableList();
    ~VariableList();
    bool Set(const char *pszName, const char *pszValue);
    char *Lookup(const char *pszName);
    Variable *GetNext(Position *ppos);

private:
    Variable *Lookup(const char *pszName, Variable ***pppvar);

private:
    Variable *m_pvarHead;
};


enum Err {
    kerrNone, kerrSyntax, kerrDivideByZero, kerrUndefinedSymbol,
    kerrNumberTooBig, kerrAsciiTooBig
};
const int kclStackMax = 5;

//
// Modules & symbols
//

struct Symbol {         // sym
    VirtualAddress va;
    char *pszName;
};

class Module {          // mod
    friend class Debugger;
    friend bool DumpSymbols(Debugger *pdbg);

public:
    Module(LocalID lidDb);
    ~Module();
    Symbol *FindSymbol(const char *pszName);
    Symbol *FindSymbol(VirtualAddress va);
    bool FindNearestSymbols(VirtualAddress va,
            char **ppszBefore, long *pcbBOffset, char **ppszAfter, 
            long *pcbAOffset);
    bool GetSymbols(VirtualAddress vaCode, long cbCode);

private:
    class Module *m_pmodNext;
    char *m_pszName;
    LocalID m_lidDb;
    long m_csym;              // count of symbols
    Symbol *m_asym;           // array of symbols
    VirtualAddress m_vaCode;
    long m_cbCode;
};


// UNDONE: Who calls this? Why? etc

class Debugger {
public:
    Debugger(bool (*pfnOutput)(const char *pszOutput));

    void Run(VirtualAddress vaStop);
    void Step();

    bool Execute(const char *pszCmd);
    bool printf(const char *pszFormat, ...);
    void DumpCpuState(VirtualAddress *pvaNextPC = NULL);
    void Disassemble(VirtualAddress vaDis, int cLines,
            VirtualAddress *pvaNextDis = NULL);

    bool SetVariable(const char *pszName, const char *pszValue);
    bool SetVariable(const char *pszName, ulong ulValue);
    bool GetVariable(const char *pszName, char *pszValue, int cchValueMax);
    bool GetVariable(const char *pszName, ulong *pulValue);
    const char *GetVariablePtr(const char *pszName);
    Variable *GetNextVariable(Position *ppos);

    Module *AddModule(LocalID lidDb);
    bool RemoveModule(LocalID lidDb);
    Module *FindModule(LocalID lidDb);

    const char *Evaluate(const char *pszExpr, long *plValue, Err *perr);
    Err EvalVariable(const char *pszName, long *plValue);
    const char *LookupSymbol(VirtualAddress va);
    VirtualAddress LookupAddress(const char *pszSymbol);
    bool FindNearestSymbols(VirtualAddress va, char **ppszBefore = NULL, 
            long *pcbBOffset = NULL, char **ppszAfter = NULL, 
            long *pcbAOffset = NULL);

    int GetDefaultBase();
    void SetDefaultBase(int nBase);

private:
    const char *EvalNumber(const char *pszExpr, long *plResult, Err *perr);

private:
    bool (*m_pfnOutput)(const char *pszOutput);
    VariableList m_vl;
    Module *m_pmodHead;
};

extern ulong GetPC();
extern ulong *GetRegPtr(CPU_regs &regs, const char *pszName);
extern int GetInstructionLength(Debugger *pdbg, VirtualAddress va);
extern bool DebugCommandHandler(Debugger *pdbg, char *apszArgs[], int cpszArgs);

#define kszNextDisasm "_NextDisasm"
#define kszNextDump "_NextDump"
#define kszNumLines "_NumLines"
#define kszBreakCmd "_BreakCmd"
#define kszAppStartup "_AppStartup"

//
// Breakpoint stuff
//

const ushort kopcBreak = 0x4AFC;
const ushort kopcRts = 0x4e75;
const ushort kopcLink = 0x4e56;

const long kcbpUserMax = 20;
const long kcbpMax = 40;

typedef long Bpid;      // bpid
const Bpid kbpidGo = kcbpUserMax;
const Bpid kbpidSysAppLaunch = kcbpUserMax + 1;
const Bpid kbpidSysAppLaunchRts = kcbpUserMax + 2;
const Bpid kbpidSysAppLaunchCallApp = kcbpUserMax + 3;
const Bpid kbpidSysAppLaunchCallReturn = kcbpUserMax + 4;
const Bpid kbpidNextAppStartup = kcbpUserMax + 5;
const Bpid kbpidAppEntrypoint = kcbpUserMax + 6;
const Bpid kbpidAll = -1;
const Bpid kbpidUnspecified = -2;

typedef bool (*Pfnbph)(Debugger *pdbg, Bpid bpid);

struct Breakpoint {     // bp
    VirtualAddress va;
    ushort ff;
    ushort usInstr;
    Pfnbph pfnbph;
};

const ushort kfBpInUse = 0x0001;
const ushort kfBpEnabled = 0x0002;

bool SetBreakpoint(VirtualAddress va, Bpid bpid = kbpidUnspecified, 
        Pfnbph pfnbph = NULL);
bool ClearBreakpoint(Bpid bpid);
bool EnableBreakpoint(Bpid bpid, bool fEnable);
void InstallBreakpoints();
void RemoveBreakpoints();
bool DoBreakpoints(Debugger *pdbg);
void ShowBreakpoints(Debugger *pdbg);

//
// Some handy helpers
//

inline VirtualAddress GetHandlePtr(ulong h)
{
    return CPU_getlong(h & 0x7fffffff);
}

inline long GetHandleSize(ulong h)
{
    VirtualAddress va = GetHandlePtr(h);
    byte flags = CPU_getbyte(va - 6 + 3);
    return (CPU_getword(va - 6) - (flags & 0xf)) - 6;
}
